﻿using System;
using System.Diagnostics;
using System.IO;
using HeuristicLab.Common;
using HeuristicLab.Core;

namespace HeuristicLab.Algorithms.DataAnalysis.SymRegGrammarEnumeration {
  class SentenceLogger : Item, IGrammarEnumerationAnalyzer {
    private readonly string distinctSentencesFileName = Environment.GetFolderPath(Environment.SpecialFolder.DesktopDirectory) + @"\distinctSentences.csv";
    private readonly string allSentencesFileName = Environment.GetFolderPath(Environment.SpecialFolder.DesktopDirectory) + @"\allSentences.csv";
    private readonly string columnDelimiter = ";";
    private readonly string header = "infix;postfix;hash";

    private TextWriterTraceListener distinctSentencesFileTrace;
    private TextWriterTraceListener allSentencesFileTrace;

    public SentenceLogger() { }

    protected SentenceLogger(SentenceLogger original, Cloner cloner) { }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new SentenceLogger(this, cloner);
    }

    public void Register(GrammarEnumerationAlgorithm algorithm) {
      algorithm.Started += GrammarEnumerationAlgorithmOnStarted;
      algorithm.Stopped += GrammarEnumerationAlgorithmOnStopped;
      algorithm.ExceptionOccurred += GrammarEnumerationAlgorithmOnStopped;

      algorithm.SentenceGenerated += SentenceGenerated;
      algorithm.DistinctSentenceGenerated += DistinctSentenceGenerated;
    }

    public void Deregister(GrammarEnumerationAlgorithm algorithm) {
      algorithm.Started -= GrammarEnumerationAlgorithmOnStarted;
      algorithm.Stopped -= GrammarEnumerationAlgorithmOnStopped;
      algorithm.ExceptionOccurred -= GrammarEnumerationAlgorithmOnStopped;

      algorithm.SentenceGenerated -= SentenceGenerated;
      algorithm.DistinctSentenceGenerated -= DistinctSentenceGenerated;
    }

    private void GrammarEnumerationAlgorithmOnStarted(object sender, EventArgs eventArgs) {
      distinctSentencesFileTrace = new TextWriterTraceListener(new FileStream(distinctSentencesFileName, FileMode.Create));
      allSentencesFileTrace = new TextWriterTraceListener(new FileStream(allSentencesFileName, FileMode.Create));
      ((StreamWriter)distinctSentencesFileTrace.Writer).AutoFlush = true;
      ((StreamWriter)allSentencesFileTrace.Writer).AutoFlush = true;

      distinctSentencesFileTrace.WriteLine(header);
      allSentencesFileTrace.WriteLine(header);
    }

    private void GrammarEnumerationAlgorithmOnStopped(object sender, EventArgs eventArgs) {
      distinctSentencesFileTrace.Close();
      allSentencesFileTrace.Close();
    }

    private void DistinctSentenceGenerated(object sender, PhraseAddedEventArgs phraseAddedEventArgs) {
      distinctSentencesFileTrace.WriteLine(ToCsvLine(
        ((GrammarEnumerationAlgorithm)sender).Grammar.ToInfixString(phraseAddedEventArgs.NewPhrase),
        phraseAddedEventArgs.NewPhrase.ToString(),
        phraseAddedEventArgs.NewHash.ToString()));
    }

    private void SentenceGenerated(object sender, PhraseAddedEventArgs phraseAddedEventArgs) {
      allSentencesFileTrace.WriteLine(ToCsvLine(
        ((GrammarEnumerationAlgorithm)sender).Grammar.ToInfixString(phraseAddedEventArgs.NewPhrase),
        phraseAddedEventArgs.NewPhrase.ToString(),
        phraseAddedEventArgs.NewHash.ToString()));
    }

    private string ToCsvLine(params string[] cols) {
      return string.Join(columnDelimiter, cols);
    }
  }
}
